@@ -17,9 +17,14 @@ class Connection;
 
 class ConnectResponse : public Response {
 public:
-    ConnectResponse(lcb_error_t err, ConnectionPtr connection);
+    ConnectResponse(
+        lcb_error_t err = LCB_SUCCESS, ConnectionPtr connection = nullptr);
 
+    std::shared_ptr<Connection> connection() const;
+
+#if !defined(NO_ERLANG)
     nifpp::TERM toTerm(const Env &env) const;
+#endif
 
 private:
     std::shared_ptr<Connection> m_connection;
