@@ -25,6 +25,7 @@ GetResponse::GetResponse(const void *key, std::size_t keySize, lcb_cas_t cas,
 {
 }
 
+#if !defined(NO_ERLANG)
 nifpp::TERM GetResponse::toTerm(const Env &env) const
 {
     if (m_err == LCB_SUCCESS) {
@@ -36,5 +37,6 @@ nifpp::TERM GetResponse::toTerm(const Env &env) const
 
     return nifpp::make(env, std::make_tuple(m_key, Response::toTerm(env)));
 }
+#endif
 
 } // namespace cb
