@@ -1,11 +1,9 @@
 #include "../nifpp.h"
+#include "helpers/cachingStorageHelperCreator.h"
 #include "helpers/init.h"
 #include "helpers/storageHelperCreator.h"
 #include "monitoring/monitoring.h"
 
-#include <asio.hpp>
-#include <asio/executor_work.hpp>
-
 #include <chrono>
 #include <map>
 #include <memory>
@@ -22,13 +20,16 @@
 #include <sys/stat.h>
 #include <sys/types.h>
 
+#include <folly/executors/IOThreadPoolExecutor.h>
+
 namespace {
 /**
  * @defgroup StaticAtoms Statically created atoms for ease of usage.
  * @{
  */
-nifpp::str_atom ok{"ok"};
-nifpp::str_atom error{"error"};
+nifpp::str_atom ok {"ok"};
+nifpp::str_atom error {"error"};
+nifpp::str_atom unknown_error {"unknown_error"};
 /** @} */
 
 using helper_ptr = one::helpers::StorageHelperPtr;
@@ -40,70 +41,83 @@ using helper_args_t = std::unordered_map<folly::fbstring, folly::fbstring>;
  * Static resource holder.
  */
 struct HelpersNIF {
-    struct HelperIOService {
-        asio::io_service service;
-        asio::executor_work<asio::io_service::executor_type> work =
-            asio::make_work(service);
-        folly::fbvector<std::thread> workers;
-    };
-
     HelpersNIF(std::unordered_map<folly::fbstring, folly::fbstring> args)
     {
         using namespace one::helpers;
 
         bufferingEnabled = (args["buffer_helpers"] == "true");
 
-        for (const auto &entry :
-            std::unordered_map<folly::fbstring, folly::fbstring>(
-                {{CEPH_HELPER_NAME, "ceph_helper_threads_number"},
-                    {POSIX_HELPER_NAME, "posix_helper_threads_number"},
-                    {S3_HELPER_NAME, "s3_helper_threads_number"},
-                    {SWIFT_HELPER_NAME, "swift_helper_threads_number"},
-                    {GLUSTERFS_HELPER_NAME,
-                        "glusterfs_helper_threads_number"}})) {
-            auto threads = std::stoul(args[entry.second].toStdString());
-            services.emplace(entry.first, std::make_unique<HelperIOService>());
-            auto &service = services[entry.first]->service;
-            auto &workers = services[entry.first]->workers;
-            for (std::size_t i = 0; i < threads; ++i) {
-                workers.push_back(std::thread([&]() { service.run(); }));
-            }
+        for (const auto &entry : std::unordered_map<folly::fbstring,
+                 std::pair<folly::fbstring, folly::fbstring>>(
+                 {{CEPH_HELPER_NAME, {"ceph_helper_threads_number", "ceph_t"}},
+                     {CEPHRADOS_HELPER_NAME,
+                         {"cephrados_helper_threads_number", "crados_t"}},
+                     {POSIX_HELPER_NAME,
+                         {"posix_helper_threads_number", "posix_t"}},
+                     {S3_HELPER_NAME, {"s3_helper_threads_number", "s3_t"}},
+                     {SWIFT_HELPER_NAME,
+                         {"swift_helper_threads_number", "swift_t"}},
+                     {GLUSTERFS_HELPER_NAME,
+                         {"glusterfs_helper_threads_number", "gluster_t"}},
+                     {WEBDAV_HELPER_NAME,
+                         {"webdav_helper_threads_number", "webdav_t"}},
+                     {XROOTD_HELPER_NAME,
+                         {"xrootd_helper_threads_number", "xrootd_t"}},
+                     {NFS_HELPER_NAME, {"nfs_helper_threads_number", "nfs_t"}},
+                     {NULL_DEVICE_HELPER_NAME,
+                         {"nulldevice_helper_threads_number", "nulldev_t"}}})) {
+            auto threadNumber =
+                std::stoul(args[entry.second.first].toStdString());
+            executors.emplace(entry.first,
+                std::make_shared<folly::IOThreadPoolExecutor>(threadNumber,
+                    std::make_shared<StorageWorkerFactory>(
+                        entry.second.second)));
         }
 
-        SHCreator = std::make_unique<one::helpers::StorageHelperCreator>(
-            services[CEPH_HELPER_NAME]->service,
-            services[POSIX_HELPER_NAME]->service,
-            services[S3_HELPER_NAME]->service,
-            services[SWIFT_HELPER_NAME]->service,
-            services[GLUSTERFS_HELPER_NAME]->service,
-            std::stoul(args["buffer_scheduler_threads_number"].toStdString()),
-            buffering::BufferLimits{
-                std::stoul(args["read_buffer_min_size"].toStdString()),
-                std::stoul(args["read_buffer_max_size"].toStdString()),
-                std::chrono::seconds{std::stoul(
-                    args["read_buffer_prefetch_duration"].toStdString())},
-                std::stoul(args["write_buffer_min_size"].toStdString()),
-                std::stoul(args["write_buffer_max_size"].toStdString()),
-                std::chrono::seconds{std::stoul(
-                    args["write_buffer_flush_delay"].toStdString())}});
+        auto storage_helper_creator =
+            std::make_unique<one::helpers::StorageHelperCreator<void>>(
+                executors[CEPH_HELPER_NAME], executors[CEPHRADOS_HELPER_NAME],
+                executors[POSIX_HELPER_NAME], executors[S3_HELPER_NAME],
+                executors[SWIFT_HELPER_NAME], executors[GLUSTERFS_HELPER_NAME],
+                executors[WEBDAV_HELPER_NAME], executors[XROOTD_HELPER_NAME],
+                executors[NFS_HELPER_NAME], executors[NULL_DEVICE_HELPER_NAME],
+                std::stoul(
+                    args["buffer_scheduler_threads_number"].toStdString()),
+                buffering::BufferLimits {
+                    std::stoul(args["read_buffer_min_size"].toStdString()),
+                    std::stoul(args["read_buffer_max_size"].toStdString()),
+                    std::chrono::seconds {std::stoul(
+                        args["read_buffer_prefetch_duration"].toStdString())},
+                    std::stoul(args["write_buffer_min_size"].toStdString()),
+                    std::stoul(args["write_buffer_max_size"].toStdString()),
+                    std::chrono::seconds {std::stoul(
+                        args["write_buffer_flush_delay"].toStdString())}});
+
+        const auto helpersCacheExpirySeconds =
+            args.count("helpers_cache_expiry_seconds") > 0
+            ? std::stoul(args.at("helpers_cache_expiry_seconds").toStdString())
+            : one::helpers::kHelperCacheDefaultExpirySeconds;
+
+        SHCreator =
+            std::make_unique<one::helpers::CachingStorageHelperCreator<void>>(
+                std::move(storage_helper_creator),
+                std::chrono::milliseconds {helpersCacheExpirySeconds * 1000});
 
         umask(0);
     }
 
     ~HelpersNIF()
     {
-        for (auto &service : services) {
-            service.second->service.stop();
-            for (auto &worker : service.second->workers) {
-                worker.join();
-            }
+        for (auto &executor : executors) {
+            executor.second->stop();
         }
     }
 
-    bool bufferingEnabled = false;
-    std::unordered_map<folly::fbstring, std::unique_ptr<HelperIOService>>
-        services;
-    std::unique_ptr<one::helpers::StorageHelperCreator> SHCreator;
+    bool bufferingEnabled {false};
+    std::unordered_map<folly::fbstring,
+        std::shared_ptr<folly::IOThreadPoolExecutor>>
+        executors;
+    std::unique_ptr<one::helpers::CachingStorageHelperCreator<void>> SHCreator;
 };
 
 std::unique_ptr<HelpersNIF> application;
@@ -115,7 +129,7 @@ namespace {
  *           POSIX open mode / flag.
  * @{
  */
-const std::unordered_map<nifpp::str_atom, one::helpers::Flag> atom_to_flag{
+const std::unordered_map<nifpp::str_atom, one::helpers::Flag> atom_to_flag {
     {"O_NONBLOCK", one::helpers::Flag::NONBLOCK},
     {"O_APPEND", one::helpers::Flag::APPEND},
     {"O_ASYNC", one::helpers::Flag::ASYNC},
@@ -143,7 +157,7 @@ one::helpers::FlagsSet translateFlags(folly::fbvector<nifpp::str_atom> atoms)
             flags.insert(result->second);
         }
         else {
-            throw std::system_error{
+            throw std::system_error {
                 std::make_error_code(std::errc::invalid_argument)};
         }
     }
@@ -246,7 +260,8 @@ std::map<std::error_code, nifpp::str_atom> error_to_atom = {
     {make_sys_error_code(std::errc::too_many_links), "emlink"},
     {make_sys_error_code(std::errc::too_many_symbolic_link_levels), "eloop"},
     {make_sys_error_code(std::errc::value_too_large), "eoverflow"},
-    {make_sys_error_code(std::errc::wrong_protocol_type), "eprototype"}};
+    {make_sys_error_code(std::errc::wrong_protocol_type), "eprototype"},
+    {make_sys_error_code(EKEYEXPIRED), "ekeyexpired"}};
 /** @} */
 
 /**
@@ -259,7 +274,7 @@ public:
      * deleter.
      */
     Env()
-        : env{enif_alloc_env(), enif_free_env}
+        : env {enif_alloc_env(), enif_free_env}
     {
     }
 
@@ -302,9 +317,32 @@ private:
     static thread_local std::default_random_engine gen;
     static thread_local std::uniform_int_distribution<int> dist;
 };
-thread_local std::random_device NifCTX::rd{};
-thread_local std::default_random_engine NifCTX::gen{NifCTX::rd()};
-thread_local std::uniform_int_distribution<int> NifCTX::dist{};
+thread_local std::random_device NifCTX::rd {};
+thread_local std::default_random_engine NifCTX::gen {NifCTX::rd()};
+thread_local std::uniform_int_distribution<int> NifCTX::dist {};
+
+/**
+ * Convert std::system_error to Erlang tuple
+ */
+std::tuple<nifpp::str_atom, nifpp::str_atom, std::string> to_error(
+    const std::system_error &e)
+{
+    auto it = error_to_atom.find(e.code());
+    nifpp::str_atom reason {e.code().message()};
+    if (it != error_to_atom.end())
+        reason = it->second;
+
+    return std::make_tuple(error, reason, e.what());
+}
+
+/**
+ * Convert std::exception to Erlang tuple
+ */
+std::tuple<nifpp::str_atom, nifpp::str_atom, std::string> to_error(
+    const std::exception &e)
+{
+    return std::make_tuple(error, unknown_error, e.what());
+}
 
 /**
  * Runs given function and returns result or error term.
@@ -318,12 +356,10 @@ template <class T> ERL_NIF_TERM handle_errors(ErlNifEnv *env, T &&fun)
         return enif_make_badarg(env);
     }
     catch (const std::system_error &e) {
-        return nifpp::make(
-            env, std::make_tuple(error, nifpp::str_atom{e.code().message()}));
+        return nifpp::make(env, to_error(e));
     }
     catch (const std::exception &e) {
-        return nifpp::make(
-            env, std::make_tuple(error, folly::fbstring{e.what()}));
+        return nifpp::make(env, to_error(e));
     }
 }
 
@@ -332,14 +368,14 @@ ERL_NIF_TERM wrap_helper(ERL_NIF_TERM (*fun)(NifCTX ctx, Args...),
     ErlNifEnv *env, const ERL_NIF_TERM args[], std::index_sequence<I...>)
 {
     return handle_errors(env,
-        [&]() { return fun(NifCTX{env}, nifpp::get<Args>(env, args[I])...); });
+        [&]() { return fun(NifCTX {env}, nifpp::get<Args>(env, args[I])...); });
 }
 
 template <typename... Args>
 ERL_NIF_TERM wrap(ERL_NIF_TERM (*fun)(NifCTX, Args...), ErlNifEnv *env,
     const ERL_NIF_TERM args[])
 {
-    return wrap_helper(fun, env, args, std::index_sequence_for<Args...>{});
+    return wrap_helper(fun, env, args, std::index_sequence_for<Args...> {});
 }
 
 template <typename... Args, std::size_t... I>
@@ -355,7 +391,7 @@ ERL_NIF_TERM noctx_wrap(ERL_NIF_TERM (*fun)(ErlNifEnv *env, Args...),
     ErlNifEnv *env, const ERL_NIF_TERM args[])
 {
     return noctx_wrap_helper(
-        fun, env, args, std::index_sequence_for<Args...>{});
+        fun, env, args, std::index_sequence_for<Args...> {});
 }
 
 /**
@@ -383,6 +419,17 @@ void handle_value(const NifCTX &ctx, file_handle_ptr handle)
     ctx.send(std::make_tuple(ok, std::move(resource)));
 }
 
+/**
+ * Handles std::pair result from helpers and sends it to requesting process.
+ */
+template <class T1, class T2>
+void handle_value(const NifCTX &ctx, std::pair<T1, T2> &&response)
+{
+    ctx.send(std::make_tuple(ok,
+        std::make_tuple(std::forward<T1>(response.first),
+            std::forward<T2>(response.second))));
+}
+
 /**
  * Handles generic result from helpers and sends it to requesting process.
  */
@@ -401,19 +448,29 @@ void handle_value(const NifCTX &ctx, folly::Unit) { ctx.send(ok); }
  */
 template <class T> void handle_result(NifCTX ctx, folly::Future<T> future)
 {
-    future.then([ctx](T &&value) { handle_value(ctx, std::move(value)); })
-        .onError([ctx](const std::system_error &e) {
-            auto it = error_to_atom.find(e.code());
-            nifpp::str_atom reason{e.code().message()};
-            if (it != error_to_atom.end())
-                reason = it->second;
+    std::move(future)
+        .thenValue(
+            [ctx](T &&value) { handle_value(ctx, std::forward<T>(value)); })
+        .thenError(folly::tag_t<std::system_error> {},
+            [ctx](auto &&e) { ctx.send(to_error(e)); })
+        .thenError(folly::tag_t<std::exception> {},
+            [ctx](auto &&e) { ctx.send(to_error(e)); });
+}
 
-            ctx.send(std::make_tuple(error, reason));
+/**
+ * Handles open file handle callback result.
+ */
+template <class T>
+void handle_result(NifCTX ctx, file_handle_ptr handle, folly::Future<T> future)
+{
+    std::move(future)
+        .thenValue([ctx, handle = std::move(handle)](T &&value) {
+            handle_value(ctx, std::forward<T>(value));
         })
-        .onError([ctx](const std::exception &e) {
-            nifpp::str_atom reason{e.what()};
-            ctx.send(std::make_tuple(error, reason));
-        });
+        .thenError(folly::tag_t<std::system_error> {},
+            [ctx](auto &&e) { ctx.send(to_error(e)); })
+        .thenError(folly::tag_t<std::exception> {},
+            [ctx](auto &&e) { ctx.send(to_error(e)); });
 }
 
 /**
@@ -449,6 +506,10 @@ static void configurePerformanceMonitoring(
             ONE_METRIC_COUNTER_SET(
                 "comp.oneprovider.mod.options.ceph_helper_thread_count",
                 std::stoul(args["ceph_helper_threads_number"].toStdString()));
+            ONE_METRIC_COUNTER_SET(
+                "comp.oneprovider.mod.options.cephrados_helper_thread_count",
+                std::stoul(
+                    args["cephrados_helper_threads_number"].toStdString()));
             ONE_METRIC_COUNTER_SET(
                 "comp.oneprovider.mod.options.posix_helper_thread_count",
                 std::stoul(args["posix_helper_threads_number"].toStdString()));
@@ -462,6 +523,16 @@ static void configurePerformanceMonitoring(
                 "comp.oneprovider.mod.options.glusterfs_helper_thread_count",
                 std::stoul(
                     args["glusterfs_helper_threads_number"].toStdString()));
+            ONE_METRIC_COUNTER_SET(
+                "comp.oneprovider.mod.options.nulldevice_helper_thread_count",
+                std::stoul(
+                    args["nulldevice_helper_threads_number"].toStdString()));
+            ONE_METRIC_COUNTER_SET(
+                "comp.oneprovider.mod.options.xrootd_helper_thread_count",
+                std::stoul(args["xrootd_helper_threads_number"].toStdString()));
+            ONE_METRIC_COUNTER_SET(
+                "comp.oneprovider.mod.options.nfs_helper_thread_count",
+                std::stoul(args["nfs_helper_threads_number"].toStdString()));
             ONE_METRIC_COUNTER_SET("comp.oneprovider.mod.options.buffer_"
                                    "scheduler_helper_thread_count",
                 std::stoul(
@@ -494,13 +565,14 @@ static void configurePerformanceMonitoring(
 }
 
 /*********************************************************************
-*
-*                          WRAPPERS (NIF based)
-*       All functions below are described in helpers_nif.erl
-*
-*********************************************************************/
+ *
+ *                          WRAPPERS (NIF based)
+ *       All functions below are described in helpers_nif.erl
+ *
+ *********************************************************************/
 
-ERL_NIF_TERM get_handle(ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
+ERL_NIF_TERM get_helper_handle(
+    ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
 {
     auto name = nifpp::get<folly::fbstring>(env, argv[0]);
     auto params = nifpp::get<helper_args_t>(env, argv[1]);
@@ -511,6 +583,37 @@ ERL_NIF_TERM get_handle(ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
     return nifpp::make(env, std::make_tuple(ok, resource));
 }
 
+ERL_NIF_TERM clean_helper_cache(
+    ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
+{
+    application->SHCreator->clean();
+    return nifpp::make(env, ok);
+}
+
+ERL_NIF_TERM get_helper_cache_stats(
+    ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
+{
+    auto stats = application->SHCreator->cacheStats();
+    return nifpp::make(env, std::make_tuple(ok, std::move(stats)));
+}
+
+ERL_NIF_TERM get_helper_id(NifCTX ctx, helper_ptr helper)
+{
+    return nifpp::make(ctx.env, std::make_tuple(ok, helper->id()));
+}
+
+ERL_NIF_TERM refresh_params(NifCTX ctx, helper_ptr helper, helper_args_t args)
+{
+    handle_result(ctx, helper->updateHelper(args));
+    return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
+}
+
+ERL_NIF_TERM check_storage_availability(NifCTX ctx, helper_ptr helper)
+{
+    handle_result(ctx, helper->checkStorageAvailability());
+    return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
+}
+
 ERL_NIF_TERM getattr(NifCTX ctx, helper_ptr helper, folly::fbstring file)
 {
     handle_result(ctx, helper->getattr(file));
@@ -524,6 +627,13 @@ ERL_NIF_TERM readdir(NifCTX ctx, helper_ptr helper, folly::fbstring file,
     return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
 }
 
+ERL_NIF_TERM listobjects(NifCTX ctx, helper_ptr helper, folly::fbstring prefix,
+    folly::fbstring marker, const off_t offset, const size_t count)
+{
+    handle_result(ctx, helper->listobjects(prefix, marker, offset, count));
+    return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
+}
+
 ERL_NIF_TERM access(
     NifCTX ctx, helper_ptr helper, folly::fbstring file, const int mask)
 {
@@ -547,9 +657,10 @@ ERL_NIF_TERM mkdir(
     return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
 }
 
-ERL_NIF_TERM unlink(NifCTX ctx, helper_ptr helper, folly::fbstring file)
+ERL_NIF_TERM unlink(NifCTX ctx, helper_ptr helper, folly::fbstring file,
+    const size_t currentSize)
 {
-    handle_result(ctx, helper->unlink(file));
+    handle_result(ctx, helper->unlink(file, currentSize));
     return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
 }
 
@@ -594,10 +705,10 @@ ERL_NIF_TERM chown(NifCTX ctx, helper_ptr helper, folly::fbstring file,
     return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
 }
 
-ERL_NIF_TERM truncate(
-    NifCTX ctx, helper_ptr helper, folly::fbstring file, const off_t size)
+ERL_NIF_TERM truncate(NifCTX ctx, helper_ptr helper, folly::fbstring file,
+    const off_t size, const size_t currentSize)
 {
-    handle_result(ctx, helper->truncate(file, size));
+    handle_result(ctx, helper->truncate(file, size, currentSize));
     return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
 }
 
@@ -629,6 +740,20 @@ ERL_NIF_TERM listxattr(NifCTX ctx, helper_ptr helper, folly::fbstring file)
     return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
 }
 
+ERL_NIF_TERM flushbuffer(
+    NifCTX ctx, helper_ptr helper, folly::fbstring file, size_t size)
+{
+    handle_result(ctx, helper->flushBuffer(file, size));
+    return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
+}
+
+ERL_NIF_TERM blocksize_for_path(
+    NifCTX ctx, helper_ptr helper, folly::fbstring file)
+{
+    handle_result(ctx, helper->blockSizeForPath(file));
+    return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
+}
+
 ERL_NIF_TERM open(NifCTX ctx, helper_ptr helper, folly::fbstring file,
     folly::fbvector<nifpp::str_atom> flags)
 {
@@ -639,32 +764,49 @@ ERL_NIF_TERM open(NifCTX ctx, helper_ptr helper, folly::fbstring file,
 
 ERL_NIF_TERM read(NifCTX ctx, file_handle_ptr handle, off_t offset, size_t size)
 {
-    handle_result(ctx, handle->read(offset, size));
+    handle_result(ctx, handle, handle->read(offset, size));
     return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
 }
 
 ERL_NIF_TERM write(NifCTX ctx, file_handle_ptr handle, const off_t offset,
-    folly::IOBufQueue buf)
+    std::pair<const uint8_t *, size_t> data)
 {
-    handle_result(ctx, handle->write(offset, std::move(buf)));
+    folly::IOBufQueue buf {folly::IOBufQueue::cacheChainLength()};
+
+    auto helperBlockSize = handle->helper()->blockSize();
+    auto bufferBlockSize =
+        (helperBlockSize != 0) ? helperBlockSize : (1U << 31);
+
+    buf.wrapBuffer(data.first, data.second, bufferBlockSize);
+
+    handle_result(ctx, handle, handle->write(offset, std::move(buf), {}));
     return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
 }
 
 ERL_NIF_TERM release(NifCTX ctx, file_handle_ptr handle)
 {
-    handle_result(ctx, handle->release());
+    handle_result(ctx, handle, handle->release());
     return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
 }
 
 ERL_NIF_TERM flush(NifCTX ctx, file_handle_ptr handle)
 {
-    handle_result(ctx, handle->flush());
+    handle_result(ctx, handle, handle->flush());
     return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
 }
 
 ERL_NIF_TERM fsync(NifCTX ctx, file_handle_ptr handle, const int isdatasync)
 {
-    handle_result(ctx, handle->fsync(isdatasync));
+    handle_result(ctx, handle, handle->fsync(isdatasync));
+    return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
+}
+
+ERL_NIF_TERM refresh_helper_params(
+    NifCTX ctx, file_handle_ptr handle, helper_args_t args)
+{
+    handle_result(ctx,
+        handle->refreshHelperParams(one::helpers::StorageHelperParams::create(
+            handle->helper()->name(), args)));
     return nifpp::make(ctx.env, std::make_tuple(ok, ctx.reqId));
 }
 
@@ -703,12 +845,36 @@ static int load(ErlNifEnv *env, void **priv_data, ERL_NIF_TERM load_info)
             env, nullptr, "file_handle_ptr"));
 }
 
+static ERL_NIF_TERM sh_refresh_params(
+    ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
+{
+    return wrap(refresh_params, env, argv);
+}
+
+static ERL_NIF_TERM sh_get_helper_id(
+    ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
+{
+    return wrap(get_helper_id, env, argv);
+}
+
+static ERL_NIF_TERM sh_listobjects(
+    ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
+{
+    return wrap(listobjects, env, argv);
+}
+
 static ERL_NIF_TERM sh_readdir(
     ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
 {
     return wrap(readdir, env, argv);
 }
 
+static ERL_NIF_TERM sh_check_storage_availability(
+    ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
+{
+    return wrap(check_storage_availability, env, argv);
+}
+
 static ERL_NIF_TERM sh_getattr(
     ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
 {
@@ -804,6 +970,18 @@ static ERL_NIF_TERM sh_listxattr(
     return wrap(listxattr, env, argv);
 }
 
+static ERL_NIF_TERM sh_flushbuffer(
+    ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
+{
+    return wrap(flushbuffer, env, argv);
+}
+
+static ERL_NIF_TERM sh_blocksize_for_path(
+    ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
+{
+    return wrap(blocksize_for_path, env, argv);
+}
+
 static ERL_NIF_TERM sh_open(ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
 {
     return wrap(open, env, argv);
@@ -814,6 +992,12 @@ static ERL_NIF_TERM sh_read(ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
     return wrap(read, env, argv);
 }
 
+static ERL_NIF_TERM sh_refresh_helper_params(
+    ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
+{
+    return wrap(refresh_helper_params, env, argv);
+}
+
 static ERL_NIF_TERM sh_write(
     ErlNifEnv *env, int argc, const ERL_NIF_TERM argv[])
 {
@@ -838,18 +1022,46 @@ static ERL_NIF_TERM sh_fsync(
     return wrap(fsync, env, argv);
 }
 
-static ErlNifFunc nif_funcs[] = {{"get_handle", 2, get_handle},
-    {"start_monitoring", 0, start_monitoring},
-    {"stop_monitoring", 0, stop_monitoring}, {"getattr", 2, sh_getattr},
-    {"access", 3, sh_access}, {"readdir", 4, sh_readdir},
-    {"mknod", 5, sh_mknod}, {"mkdir", 3, sh_mkdir}, {"unlink", 2, sh_unlink},
-    {"rmdir", 2, sh_rmdir}, {"symlink", 3, sh_symlink},
-    {"rename", 3, sh_rename}, {"link", 3, sh_link}, {"chmod", 3, sh_chmod},
-    {"chown", 4, sh_chown}, {"truncate", 3, sh_truncate},
-    {"setxattr", 6, sh_setxattr}, {"getxattr", 3, sh_getxattr},
-    {"removexattr", 3, sh_removexattr}, {"listxattr", 2, sh_listxattr},
-    {"open", 3, sh_open}, {"read", 3, sh_read}, {"write", 3, sh_write},
-    {"release", 1, sh_release}, {"flush", 1, sh_flush}, {"fsync", 2, sh_fsync}};
+static ErlNifFunc nif_funcs[] = {
+    {"get_helper_handle", 2, get_helper_handle, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"get_helper_id", 1, sh_get_helper_id, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"start_monitoring", 0, start_monitoring, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"stop_monitoring", 0, stop_monitoring, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"refresh_params", 2, sh_refresh_params, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"refresh_helper_params", 2, sh_refresh_helper_params,
+        ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"check_storage_availability", 1, sh_check_storage_availability,
+        ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"getattr", 2, sh_getattr, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"access", 3, sh_access, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"readdir", 4, sh_readdir, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"listobjects", 5, sh_listobjects, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"mknod", 5, sh_mknod, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"mkdir", 3, sh_mkdir, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"unlink", 3, sh_unlink, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"rmdir", 2, sh_rmdir, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"symlink", 3, sh_symlink, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"rename", 3, sh_rename, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"link", 3, sh_link, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"chmod", 3, sh_chmod, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"chown", 4, sh_chown, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"truncate", 4, sh_truncate, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"setxattr", 6, sh_setxattr, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"getxattr", 3, sh_getxattr, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"removexattr", 3, sh_removexattr, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"listxattr", 2, sh_listxattr, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"open", 3, sh_open, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"read", 3, sh_read, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"write", 3, sh_write, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"release", 1, sh_release, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"flush", 1, sh_flush, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"fsync", 2, sh_fsync, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"flushbuffer", 3, sh_flushbuffer, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"blocksize_for_path", 2, sh_blocksize_for_path,
+        ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"clean_helper_cache", 0, clean_helper_cache, ERL_NIF_DIRTY_JOB_IO_BOUND},
+    {"get_helper_cache_stats", 0, get_helper_cache_stats,
+        ERL_NIF_DIRTY_JOB_IO_BOUND}};
 
 ERL_NIF_INIT(helpers_nif, nif_funcs, load, NULL, NULL, NULL);
 
